
{*******************************************************}
{                                                       }
{       TRVRuler: ruler for TRichViewEdit               }
{       Inherited from TRuler by Pieter Zijlstra.       }
{       Created by Sergey Tkachenko, basing on example  }
{       code by Pieter Zijlstra.                        }
{                                                       }
{       Sergey Tkachenko                                }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{       Pieter Zijlstra                                 }
{       p.zylstra@hccnet.nl                             }
{       http://home.hccnet.nl/p.zylstra/                }
{                                                       }
{*******************************************************}

{ --------------------------------------------------------------------------
  v1.1.1
  v1.4.0 - changed LeftInset to Inset
  v1.5.0 - added TableEditor
  v1.6.0 - added tabs
  --------------------------------------------------------------------------
  What does this ruler change:
  - Left, first, right indents of the selected paragraphs
  - RichViewEdit.LeftMargin, .RightMargin. Note: they are not paper margins,
    they are "padding" of printable area.
  - If rvoClientTextWidth is not in RichViewEdit.Options, it changes
    RichViewEdit.MaxTextWidth and RichViewEdit.MinTextWidth. Note: they do
    not affect printing.
  --------------------------------------------------------------------------
  Not implemented yet:
  - working with paragraph bullets and numbering
  - working inside table cells
  Working with TDBRichViewEdit is not tested.
  --------------------------------------------------------------------------

  Properties:
  RichViewEdit: TCustomRichViewEdit - attached editor;
  LineStyle: TPenStyle - style of lines.

  Methods:
  UpdateRulerIndents - updates ruler's indents from RichViewEdit.
  Rarely needs to be called.

  UpdateRulerMargins - updates ruler's margins from RichViewEdit
  Must be called when margins of editor were changed, for example, after
  reading RVF file.

  --------------------------------------------------------------------------

  Technical notes:

  This component changes the following events of the assigned RichViewEdit:
  OnHScrolled, OnResize, OnCurParaStyleChanged.
  Its handlers for these events call old handlers. But you must not change
  these events at run time after Ruler.RichViewEdit has been assigned.
}


unit RVRuler;

interface
uses Windows, Classes, Graphics, Controls,
     Ruler,
     RVStyle, RVScroll, RichView, RVEdit,
     RVItem, RVTable;

type
  TRVRuler = class(TRuler)
  private
    FControlCanvas: TControlCanvas;
    FFirstIndent, FLeftIndent, FRightIndent: Integer;
    FLineStyle: TPenStyle;
    FRichViewEdit: TCustomRichViewEdit;
    FRVEOldHScrolled, FRVEOldResize, FRVEOldCurParaStyleChanged: TNotifyEvent;
    FSaveX: Integer;
    FUpdating, FUpdatingTabs: Boolean;
    FLeftInset: Integer;
    procedure AssignRichViewEditEvents;
    procedure DrawMarkLine;
    procedure RichViewEditCurParaStyleChanged(Sender: TObject);
    procedure RichViewEditHScrolled(Sender: TObject);
    procedure RichViewEditParaStyleConversion(Sender: TCustomRichViewEdit;
      StyleNo, UserData: Integer; AppliedToText: Boolean; var NewStyleNo: Integer);
    procedure RichViewEditResize(Sender: TObject);
    procedure RestoreRichViewEditEvents;
    procedure SetRichViewEdit(const Value: TCustomRichViewEdit);
  protected
    procedure DoBiDiModeChanged; override;
    procedure DoIndentChanged; override;
    procedure DoMarginChanged; override;
    procedure DoTabChanged; override;
    procedure DoRulerItemMove(X: Integer; Removing: Boolean); override;
    procedure DoRulerItemRelease; override;
    procedure DoRulerItemSelect(X: Integer); override;
    procedure DoTableColumnChanged; override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AdjustPosition;
    procedure UpdateRulerIndents;
    procedure UpdateRulerMargins;
    procedure UpdateTableEditor;
  published
    property Inset default 0;
    property LineStyle: TPenStyle read FLineStyle write FLineStyle default psDot;
    property RichViewEdit: TCustomRichViewEdit read FRichViewEdit write SetRichViewEdit;
    property TableEditor;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('RichView', [TRVRuler]);
end;

{=============================== TRVRuler =====================================}
constructor TRVRuler.Create(AOwner: TComponent);
begin
  inherited;
  Inset := 0;
  FLineStyle := psDot;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (Operation=opRemove) and (AComponent=FRichViewEdit) then begin
    RestoreRichViewEditEvents;
    FRichViewEdit := nil;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.Loaded;
begin
  inherited;
  AssignRichViewEditEvents;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoBiDiModeChanged;
begin
  inherited;
  AdjustPosition;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoRulerItemSelect(X: Integer);
begin
  inherited;
  if not Assigned(RichViewEdit) then
    Exit;
  RichViewEdit.Update;
  if not Assigned(FControlCanvas) then
    FControlCanvas := TControlCanvas.Create;
  with FControlCanvas do begin
    Control := RichViewEdit;
    Pen.Color := clBlack;
    Pen.Mode := pmXor;
    Pen.Style := LineStyle;
  end;
  FSaveX := X - FLeftInset;
  DrawMarkLine;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoRulerItemMove(X: Integer; Removing: Boolean);
begin
  inherited;
  if not Assigned(RichViewEdit) then
    Exit;
  DrawMarkLine;
  if Removing then
    FSaveX := -1
  else
    FSaveX := X - FLeftInset;
  DrawMarkLine;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoRulerItemRelease;
begin
  inherited;
  DrawMarkLine;
  FControlCanvas.Free;
  FControlCanvas := nil;
  FRichViewEdit.Refresh;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DrawMarkLine;
begin
  if Assigned(FControlCanvas) then
    with FControlCanvas do
      if FSaveX >= 0 then begin
        MoveTo(FSaveX-FRichViewEdit.HScrollPos, 0);
        LineTo(FSaveX-FRichViewEdit.HScrollPos, RichViewEdit.Height);
      end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoMarginChanged;
begin
  inherited;
  if not Assigned(FRichViewEdit) or FUpdating then
    Exit;
  if (rvfoSaveLayout in FRichViewEdit.RVFOptions) and
     not FRichViewEdit.CanChange then
    Exit;
  FRichViewEdit.LeftMargin := Round(UnitsToPixs(LeftMargin));
  FRichViewEdit.RightMargin := Round(UnitsToPixs(RightMargin));
  if not (rvoClientTextWidth in FRichViewEdit.Options) then begin
    FRichViewEdit.MinTextWidth :=
      Round(UnitsToPixs(PageWidth) - FRichViewEdit.LeftMargin - FRichViewEdit.RightMargin);
    FRichViewEdit.MaxTextWidth := FRichViewEdit.MinTextWidth;
  end;
  FRichViewEdit.RVData.Format_(True, True, False, 0, FRichViewEdit.Canvas, False,
    False);
  if rvfoSaveLayout in FRichViewEdit.RVFOptions then
    FRichViewEdit.Change;
  UpdateRulerMargins;
  UpdateTableEditor;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoTabChanged;
begin
  FUpdatingTabs := True;
  try
    DoIndentChanged
  finally
    FUpdatingTabs := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoIndentChanged;
var M: Extended;
    OldStyleConversionEvent: TRVStyleConversionEvent;
    RVEditor: TCustomRichViewEdit;
begin
  inherited;
  if not Assigned(FRichViewEdit) or FUpdating then
    Exit;

  RVEditor := FRichViewEdit.TopLevelEditor;
  M := MultiPixels; // TRichView is based on pixels
  if UseRTL then
  begin
    FFirstIndent := Round(M * FirstIndent);
    FRightIndent := Round(M * LeftIndent + FFirstIndent);
    FFirstIndent := FFirstIndent - FRightIndent;
    FLeftIndent  := Round(M * RightIndent);
  end
  else
  begin
    FFirstIndent := Round(M * FirstIndent);
    FLeftIndent  := Round(M * LeftIndent + FFirstIndent);
    FFirstIndent := FFirstIndent - FLeftIndent;
    FRightIndent := Round(M * RightIndent);
  end;

  OldStyleConversionEvent := RVEditor.OnParaStyleConversion;
  try
    RVEditor.OnParaStyleConversion := RichViewEditParaStyleConversion;
    RVEditor.ApplyParaStyleConversion(0);
  finally
    RVEditor.OnParaStyleConversion := OldStyleConversionEvent;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.SetRichViewEdit(const Value: TCustomRichViewEdit);
begin
  RestoreRichViewEditEvents;
  FRichViewEdit := Value;
  if not (csLoading in ComponentState) then
    AssignRichViewEditEvents;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.AssignRichViewEditEvents;
begin
  if not Assigned(FRichViewEdit) or (csDesigning in ComponentState) then
    Exit;
  FRVEOldResize := TRichViewEdit(FRichViewEdit).OnResize;
  FRVEOldHScrolled := TRichViewEdit(FRichViewEdit).OnHScrolled;
  FRVEOldCurParaStyleChanged := TRichViewEdit(FRichViewEdit).OnCurParaStyleChanged;
  TRichViewEdit(FRichViewEdit).OnResize := RichViewEditResize;
  TRichViewEdit(FRichViewEdit).OnHScrolled := RichViewEditHScrolled;
  TRichViewEdit(FRichViewEdit).OnCurParaStyleChanged := RichViewEditCurParaStyleChanged;
  if Assigned(FRichViewEdit.Style) then
    DefaultTabWidth := FRichViewEdit.Style.DefTabWidth/MultiPixels;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.RestoreRichViewEditEvents;
begin
  if not Assigned(FRichViewEdit) or (csDesigning in ComponentState) then
    Exit;
  TRichViewEdit(FRichViewEdit).OnResize :=  FRVEOldResize;
  TRichViewEdit(FRichViewEdit).OnHScrolled := FRVEOldHScrolled;
  TRichViewEdit(FRichViewEdit).OnCurParaStyleChanged := FRVEOldCurParaStyleChanged;
  FRVEOldResize := nil;
  FRVEOldHScrolled := nil;
  FRVEOldCurParaStyleChanged := nil;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.AdjustPosition;
var pt: TPoint;
begin
  if not Assigned(FRichViewEdit) then
    Exit;
  pt := FRichViewEdit.ClientOrigin;
  pt := FRichViewEdit.Parent.ScreenToClient(pt);
  Dec(pt.X, FRichViewEdit.Left);
  FLeftInset := -(FRichViewEdit.HScrollPos - pt.X);
  if UseRTL then
    with FRichViewEdit do
      Inset := Pred(Width - (pt.X + RVData.TextWidth + LeftMargin +
        RightMargin) + HScrollPos)
  else
    Inset := FLeftInset;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.RichViewEditHScrolled(Sender: TObject);
begin
  AdjustPosition;
  if Assigned(FRVEOldHScrolled) then
    FRVEOldHScrolled(Sender);
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.RichViewEditResize(Sender: TObject);
begin
  AdjustPosition;
  UpdateRulerMargins;
  UpdateTableEditor;
  if Assigned(FRVEOldResize) then
    FRVEOldResize(Sender);
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.RichViewEditCurParaStyleChanged(Sender: TObject);
var RVBiDiMode: TRVBiDiMode;
begin
  if Assigned(FRVEOldCurParaStyleChanged) then
    FRVEOldCurParaStyleChanged(Sender);

  RVBiDiMode := FRichViewEdit.Style.ParaStyles[FRichViewEdit.CurParaStyleNo].BiDiMode;
  if RVBiDiMode = rvbdUnspecified then
    RVBiDiMode := FRichViewEdit.BiDiMode;
  if RVBiDiMode = rvbdRightToLeft then
    BiDiModeRuler := bmRightToLeft
  else
    BiDiModeRuler := bmUseBiDiMode;

  UpdateTableEditor;
  UpdateRulerIndents;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.RichViewEditParaStyleConversion(
  Sender: TCustomRichViewEdit; StyleNo, UserData: Integer;
  AppliedToText: Boolean; var NewStyleNo: Integer);
var ParaInfo: TParaInfo;
    {$IFNDEF RVDONOTUSETABS}
    i: Integer;
    M: Extended;
    {$ENDIF}
begin
  ParaInfo := TParaInfo.Create(nil);
  try
    ParaInfo.Assign(Sender.Style.ParaStyles[StyleNo]);
    ParaInfo.LeftIndent  := FLeftIndent;
    ParaInfo.FirstIndent := FFirstIndent;
    ParaInfo.RightIndent := FRightIndent;
    {$IFNDEF RVDONOTUSETABS}
    M := MultiPixels;
    ParaInfo.Tabs.Clear;
    for i := 0 to Tabs.Count-1 do
      with ParaInfo.Tabs.Add do begin
        Position := Round(M*Tabs[i].Position);
        case Tabs[i].Align of
          taLeftAlign:   Align := rvtaLeft;
          taRightAlign:  Align := rvtaRight;
          taCenterAlign: Align := rvtaCenter;
        end;
      end;
    {$ENDIF}
    NewStyleNo := Sender.Style.ParaStyles.FindSuchStyle(StyleNo,ParaInfo,RVAllParaInfoProperties);
    if NewStyleNo=-1 then begin
      Sender.Style.ParaStyles.Add;
      NewStyleNo := Sender.Style.ParaStyles.Count-1;
      Sender.Style.ParaStyles[NewStyleNo].Assign(ParaInfo);
      Sender.Style.ParaStyles[NewStyleNo].Standard := False;
    end;
  finally
    ParaInfo.Free;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.UpdateRulerIndents;
var M: Extended;
    ParaNo: Integer;
    ParaStyles: TParaInfos;
    {$IFNDEF RVDONOTUSETABS}
    i: Integer;
    {$ENDIF}
begin
  if not Assigned(FRichViewEdit) or FUpdating or FUpdatingTabs then
    Exit;
  FUpdating := True;
  try
    ParaStyles := FRichViewEdit.Style.ParaStyles;
    ParaNo := FRichViewEdit.CurParaStyleNo;
    M := 1 / MultiPixels;
    if UseRTL then
    begin
      FirstIndent :=
        M * (ParaStyles[ParaNo].RightIndent + ParaStyles[ParaNo].FirstIndent);
      LeftIndent := M * -ParaStyles[ParaNo].FirstIndent;
      RightIndent := M * ParaStyles[ParaNo].LeftIndent;
    end
    else
    begin
      FirstIndent :=
        M * (ParaStyles[ParaNo].LeftIndent + ParaStyles[ParaNo].FirstIndent);
      LeftIndent  := M * -ParaStyles[ParaNo].FirstIndent;
      RightIndent := M * ParaStyles[ParaNo].RightIndent;
    end;

    {$IFNDEF RVDONOTUSETABS}
    Tabs.Clear;
    for i := 0 to ParaStyles[ParaNo].Tabs.Count-1 do
      with Tabs.Add do begin
        Position := M*ParaStyles[ParaNo].Tabs[i].Position;
        case ParaStyles[ParaNo].Tabs[i].Align of
          rvtaLeft:   Align := taLeftAlign;
          rvtaRight:  Align := taRightAlign;
          rvtaCenter: Align := taCenterAlign;
        end;
      end;
    {$ENDIF}
  finally
    FUpdating := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.UpdateRulerMargins;
var M: Extended;
begin
  if not Assigned(FRichViewEdit) or FUpdating then
    Exit;
  FUpdating := True;
  try
    M := 1 / MultiPixels;
    LeftMargin := M * FRichViewEdit.LeftMargin;
    RightMargin := M * FRichViewEdit.RightMargin;
    PageWidth := M * (FRichViewEdit.RVData.TextWidth + FRichViewEdit.LeftMargin +
      FRichViewEdit.RightMargin);
  finally
    FUpdating := False;
  end;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.DoTableColumnChanged;
var
  C, TW, CW: Integer;
  Data: Integer;
  M: Extended;
  ItemNo: Integer;
  RVEditor: TCustomRichViewEdit;
  RVItem: TCustomRVItemInfo;
  RVTable: TRVTableItemInfo;
begin
  inherited;
  if not Assigned(FRichViewEdit) or FUpdating then
    Exit;

  M := MultiPixels;
  if TableEditor.Active and (TableEditor.DraggedDelta <> 0) then
  begin
    if FRichViewEdit.GetCurrentItemEx(TRVTableItemInfo, RVEditor, RVItem) then
    begin
      RVTable := TRVTableItemInfo(RVItem);
      ItemNo := RVEditor.GetItemNo(RVTable);
      RVEditor.BeginItemModify(ItemNo, Data);
      RVEditor.BeginUndoGroup(rvutModifyItem);
      RVEditor.SetUndoGroupMode(True);
      try
        TW := Round(M * TableEditor.TableWidth) + 2 * RVTable.BorderWidth;
        if RVTable.BestWidth < 0 then
          RVTable.BestWidth := Pred(-100 * TW div FRichViewEdit.RVData.TextWidth)
        else
          RVTable.BestWidth := TW;
        with TableEditor do
        begin
          C := DraggedColumn;
          CW := Round(M * (Cells[C].DragBoundary + DraggedDelta +
            CellBorderWidth + CellHSpacing / 2));
          RVTable.ResizeCol(C, CW, DraggedWithShift);
        end;
      finally
        RVEditor.SetUndoGroupMode(False)
      end;
      RVEditor.EndItemModify(ItemNo, Data);
      RVEditor.Change;
    end;
  end;
  UpdateTableEditor;
end;
{------------------------------------------------------------------------------}
procedure TRVRuler.UpdateTableEditor;
var
  R, C, L, T, X, Y: Integer;
  Row, Col: Integer;
  MainRow, MainCol: Integer;
  M: Extended;
  ParaNo: Integer;
  ParaStyles: TParaInfos;
  RVCell: TRVTableCellData;
  RVCellLast: TRVTableCellData;
  RVEditor: TCustomRichViewEdit;
  RVItem: TCustomRVItemInfo;
  RVTable: TRVTableItemInfo;
  W: Extended;
begin
  if not Assigned(FRichViewEdit) then
    Exit;

  FUpdating := True;
  try
    M := 1 / MultiPixels;
    TableEditor.Active := FRichViewEdit.InplaceEditor <> nil;
    if TableEditor.Active then
    begin
      TableEditor.Cells.Clear;
      if FRichViewEdit.GetCurrentItemEx(TRVTableItemInfo, RVEditor, RVItem) then
      begin
        RVTable := TRVTableItemInfo(RVItem);
        if RVTable.GetEditedCell(Row, Col) <> nil then
        begin
          TableEditor.CellIndex       := Col;
          TableEditor.BorderHSpacing  := M * RVTable.BorderHSpacing;
          TableEditor.BorderWidth     := M * RVTable.BorderWidth;
          TableEditor.CellBorderWidth := M * RVTable.CellBorderWidth;
          TableEditor.CellHSpacing    := M * RVTable.CellHSpacing;
          TableEditor.CellPadding     := M * RVTable.CellPadding;

          // Setup columns
          RVCellLast := nil;
          for C := 0 to RVTable.Rows[Row].Count - 1 do
          begin
            RVCell := RVTable.Cells[Row, C];
            if RVCell <> nil then
            begin
              TableEditor.Cells.Add.CellWidth := M * RVCell.Width;
              RVCellLast := RVCell;
            end
            else
            begin
              RVCell := RVTable.Rows.GetMainCell(Row, C, MainRow, MainCol);
              if RVCellLast <> RVCell then
              begin
                TableEditor.Cells.Add.CellWidth := M * RVCell.Width;
                RVCellLast := RVCell;
              end
              else
                TableEditor.Cells.Add.CellWidth := -1;
            end;
          end;

          // Setup Drag boundaries
          TableEditor.UseDragBoundaries := True;
          for C := 0 to RVTable.Rows[0].Count - 1 do
            if TableEditor.Cells[C].DragBoundary <= 0 then
              for R := 0 to RVTable.Rows.Count - 1 do
                if (RVTable.Cells[R, C] <> nil) and (RVTable.Cells[R, C].ColSpan = 1) then
                begin
                  TableEditor.Cells[C].DragBoundary := M * RVTable.Cells[R, C].Width;
                  Break;
                end;
          for C := 0 to RVTable.Rows[0].Count - 1 do
            if TableEditor.Cells[C].DragBoundary = -1 then
              for R := 0 to RVTable.Rows.Count - 1 do
                if (RVTable.Cells[R, C] = nil) then
                begin
                  W := M * RVTable.Rows.GetMainCell(R, C, MainRow, MainCol).Width;
                  with TableEditor do
                  begin
                    for MainCol := MainCol to C - 1 do
                      W := W - Cells[MainCol].DragBoundary - 2 * CellBorderWidth -
                        CellHSpacing;
                    Cells[C].DragBoundary := W;
                  end;
                  Break;
                end;

          // Setup column indents
          ParaStyles := FRichViewEdit.Style.ParaStyles;
          ParaNo := FRichViewEdit.CurParaStyleNo;
          if UseRTL then
          begin
            TableEditor.FirstIndent :=
              M * (ParaStyles[ParaNo].RightIndent + ParaStyles[ParaNo].FirstIndent);
            TableEditor.LeftIndent := M * -ParaStyles[ParaNo].FirstIndent;
            TableEditor.RightIndent := M * ParaStyles[ParaNo].LeftIndent;
          end
          else
          begin
            TableEditor.FirstIndent :=
              M * (ParaStyles[ParaNo].LeftIndent + ParaStyles[ParaNo].FirstIndent);
            TableEditor.LeftIndent := M * -ParaStyles[ParaNo].FirstIndent;
            TableEditor.RightIndent := M * ParaStyles[ParaNo].RightIndent;
          end;

          // Setup table position
          if RVEditor.GetItemCoords(RVTable.GetMyItemNo, L, T) then
          begin
            RVEditor.RVData.GetOriginEx(X, Y);
            Inc(L, X);
            Inc(T, Y);
          end
          else
            L := 0;
          if TableEditor.TablePosition = tpFromMargin then
            L := L - FRichViewEdit.LeftMargin;
          TableEditor.TableLeft := M * (L + RVTable.BorderWidth);
          TableEditor.TableWidth := M * (RVTable.Width - 2 * RVTable.BorderWidth);
        end
        else
          TableEditor.Active := False;
      end
      else
        TableEditor.Active := False;
    end;
  finally
    FUpdating := False;
  end;
end;

end.
